from collections.abc import MutableMapping
from logging.config import ConvertingDict, ConvertingMixin
from typing import Any, ItemsView, Iterator, Optional, ValuesView


# pylint: disable=too-many-ancestors
class ConvertingMapping(MutableMapping, ConvertingMixin):
    """Similar to ConvertingDict without inheriting from dict, to allow **kwargs unpacking override.

    When using **kwargs, __getitem__ is called on the unpacked object unless it is a dict instance, due to an internal
    optimization.
    To convert the dict values on keyword unpacking, we have to use a class that does not inherit from dict.
    Copies all ConvertingDict methods to be able to replace it seamlessly in functions that expect ConvertingDict
    instances.
    """

    def __init__(self, converting_dict: ConvertingDict):
        self._dict = converting_dict

    __doc__ = dict.__doc__

    __hash__ = dict.__hash__

    def __getattr__(self, item: Any) -> Any:
        """Access methods from ConvertingDict such as update(), keys(), convert()..."""

        return getattr(self._dict, item)

    def get(self, key: Any, default: Any = None) -> Optional[Any]:
        return self._dict.get(key, default)

    def pop(self, key: Any, default: Any = None) -> Any:
        return self._dict.pop(key, default)

    def values(self) -> ValuesView[Any]:
        for key, value in self._dict.items():
            self.convert_with_key(key, value)
        return self._dict.values()

    def items(self) -> ItemsView[Any, Any]:
        for key, value in self._dict.items():
            self.convert_with_key(key, value)
        return self._dict.items()

    def __getitem__(self, key: Any) -> Any:
        return self._dict.__getitem__(key)

    def __setitem__(self, key: Any, value: Any) -> None:
        return self._dict.__setitem__(key, value)

    def __delitem__(self, key: Any) -> None:
        return self._dict.__delitem__(key)

    def __eq__(self, other: Any) -> bool:
        return self._dict == other

    def __ne__(self, other: Any) -> bool:
        return self._dict != other

    def __iter__(self) -> Iterator[Any]:
        return self._dict.__iter__()

    def __len__(self) -> int:
        return self._dict.__len__()

    def __contains__(self, item: Any) -> bool:
        return self._dict.__contains__(item)



from typing import Any, Dict

from configue.yaml_loader import YamlLoader
from .config_loader import ConfigLoader


def load_config_from_dict(config_dict: Dict) -> Any:
    """Load configuration from a dictionary.

    :param config_dict: the dictionary to parse
    :return: the converting dict corresponding to config_dict.
    """

    return ConfigLoader(config_dict).config


def load_config_from_file(file_path: str) -> Any:
    """Load configuration from a YAML file.

    :param file_path: Absolute path to the YAML file containing the configuration
    :return: the converting dict corresponding to the file.
    """

    config = YamlLoader(file_path).load()
    return load_config_from_dict(config)



import logging
import os
import re
from typing import Any

import yaml


class YamlLoader:
    logger = logging.getLogger(__name__)

    def __init__(self, file_path: str) -> None:
        self._file_path = file_path
        # Matches "${myvar-default}" -> "${", "myvar", "-", "default", "}"
        # or "${myvar}" -> "${", "myvar", "", "", "}"
        self._env_pattern_regex = re.compile(r"(?:(?!\${[^-}]+}).)*(\${)([^-}]+)(-?)([^}]*)(})")
        # Custom loader to bind our constructors without impacting the yaml class or the other instances of YamlLoader
        self._loader = type("Loader", (yaml.FullLoader,), {})

    def load(self) -> Any:
        yaml.add_implicit_resolver("!env", self._env_pattern_regex, Loader=self._loader)
        yaml.add_constructor("!env", self._env_constructor, Loader=self._loader)
        yaml.add_constructor("!path", self._path_constructor, Loader=self._loader)
        yaml.add_constructor("!import", self._import_constructor, Loader=self._loader)
        yaml.add_constructor("!list", self._list_constructor, Loader=self._loader)

        with open(self._file_path, encoding="utf-8") as config_file:
            return yaml.load(config_file, self._loader)

    def _env_constructor(self, loader: yaml.Loader, node: yaml.ScalarNode) -> Any:
        raw_value: str = loader.construct_scalar(node)
        replaced_str = ""
        end_pos = 0
        for match in self._env_pattern_regex.finditer(raw_value):
            env_var_name, has_default, default_value = match.group(2, 3, 4)
            start_pos = match.start(1)
            if env_var_name not in os.environ and not has_default:
                self.logger.warning(f"Missing environment var: '{env_var_name}', no default is set")
            replaced_str += f"{raw_value[end_pos:start_pos]}{os.environ.get(env_var_name, default_value)}"
            end_pos = match.end(5)
        replaced_str += raw_value[end_pos:]
        # Put back quotes
        if node.style == "\"":
            replaced_str = f"\"{replaced_str}\""
        # reload node to cast the value
        return self._get_reloaded_value(replaced_str)

    def _path_constructor(self, loader: yaml.Loader, node: yaml.ScalarNode) -> Any:
        raw_value = loader.construct_scalar(node)
        # reload node to eval environment variables
        new_value = str(self._get_reloaded_value(raw_value))
        expanded_new_value = os.path.expanduser(new_value)
        full_path = os.path.join(os.path.dirname(self._file_path), expanded_new_value)
        return full_path

    def _import_constructor(self, loader: yaml.Loader, node: yaml.ScalarNode) -> Any:
        raw_value = loader.construct_scalar(node)
        # reload node to eval environment variables
        templated_value = str(self._get_reloaded_value(raw_value))
        imported_path = os.path.join(os.path.dirname(self._file_path), templated_value)
        loader = YamlLoader(imported_path)
        return loader.load()

    def _list_constructor(self, loader: yaml.Loader, node: yaml.ScalarNode) -> Any:
        raw_value = loader.construct_scalar(node)
        # reload node to eval environment variables
        new_value = str(self._get_reloaded_value(raw_value))
        split_value = new_value.split(",")
        return split_value

    def _get_reloaded_value(self, str_value: str) -> Any:
        # Reload the value
        return yaml.load(str_value, Loader=self._loader)



import logging
from logging.config import BaseConfigurator, ConvertingDict, ConvertingList, ConvertingTuple
from typing import Any, Dict, Tuple, Optional

from configue.converting_mapping import ConvertingMapping


class ConfigLoader(BaseConfigurator):
    """A configurator based on logging.config.

    Can instantiate classes from a configuration dictionary.
    The instantiation happens when the object is fetched from the configuration.

    :ivar config: The parsed configuration.
    """

    logger = logging.getLogger(__name__)

    def __init__(self, config_dict: Dict) -> None:
        self._config: Optional[ConvertingDict] = None
        BaseConfigurator.__init__(self, config_dict)
        self.config = self.convert(self.config)

    @property
    def config(self) -> Any:
        return self._config

    @config.setter
    def config(self, config: ConvertingDict) -> None:
        self._config = config

    def convert(self, value: Any) -> Any:
        value = BaseConfigurator.convert(self, value)
        if isinstance(value, ConvertingDict):
            value = ConvertingMapping(value)
            if "()" in value:
                cls_path = value["()"]
                try:
                    return self.configure_custom(value)
                except Exception as error:
                    self.logger.error(f"Could not instantiate {cls_path}: {error!r}")
                    raise
        elif isinstance(value, ConvertingList):
            value = list(map(self.convert, value))
        elif isinstance(value, ConvertingTuple):
            value = tuple(map(self.convert, value))
        return value

    def cfg_convert(self, value: str) -> Any:
        """Default converter for the cfg:// protocol."""

        str_to_parse = value
        match = self.WORD_PATTERN.match(str_to_parse)
        if match is None:
            raise ValueError(f"Unable to convert {value!r}")
        str_to_parse = str_to_parse[match.end():]
        parsed_object = self.config[match.groups()[0]]
        while str_to_parse:
            parsed_object, str_to_parse = self._parse_next_cfg_level(parsed_object, str_to_parse)
        return parsed_object

    def _parse_next_cfg_level(self, parsed_object: Any, str_to_parse: str) -> Tuple[Any, str]:
        match = self.DOT_PATTERN.match(str_to_parse)  # parsed_object.property
        if match:
            remaining_str: str = str_to_parse[match.end():]
            if not isinstance(parsed_object, (dict, ConvertingMapping)):
                return getattr(parsed_object, match.groups()[0]), remaining_str
            return parsed_object[match.groups()[0]], remaining_str
        match = self.INDEX_PATTERN.match(str_to_parse)  # parsed_object[property]
        if not match:
            raise ValueError(f"Unable to convert {str_to_parse!r}")
        remaining_str = str_to_parse[match.end():]
        idx = match.groups()[0]
        if not self.DIGIT_PATTERN.match(idx):  # parsed_object["property"]
            return parsed_object[idx], remaining_str
        return parsed_object[int(idx)], remaining_str  # parsed_object[0]
